#include <math.h>
#include <stdio.h>
#include <string.h>

#ifndef M_PI
#define M_PI  3.14159265358979323
#endif

#define R2D(rad)   ((rad) * (180/M_PI))

/*
 * Convert between geographic position (latitude/longitude) and names of
 * facets in the triangulation of the globe.
 * To compile:
 *   cc -o subdiv subdiv.c -lm
 *
 * Written in 1994 by Stuart Levy, Geometry Center, University of Minnesota,
 * (612)624-1867, for Skyline Displays.
 *
 * Routines of interest:
 * double dot(v1, v2) Vec *v1, *v2;
 * void cross(v1, v2, result) Vec *v1, *v2, *result;
 *	Vector arithmetic.
 * double distance(v1, v2) Vec *v1, *v2;
 *	Great-circle distance between two points.
 * double normalize(v) Vec *v;
 *	Normalize a vector to lie on the unit sphere.
 * void interp(v1, v2, v3, result) Vec *v1, *v2, *v3, *result;
 *	Linear interpolation among three vectors.  Returns normalized result.
 * int rightside(v, va, vb) Vec *v, *va, *vb;
 *	On which side of the line (va,vb) does point v lie?
 * double rightby(v, va, vb, movedv) Vec *v, *va, *vb, *movedv;
 *	How far to the left of (va,vb) does point v lie?
 *	If it does lie to the left, set *movedv to nearest point on the line.
 * int within(v, va, vb, vc)
 *	Does point v lie within triangle (va, vb, vc)?
 * double withinby(v, va, vb, vc) Vec *v, *va, *vb, *vc;
 *	By how much does v lie within (va,vb,vc)?
 *
 * Subdivision routines:
 * subtri(tri, facet, subt) Vec tri[3]; int facet; Vec subt[3];
 *	Choose the facet'th subdivision of triangle "tri"; vertices into "subt".
 * findtri(t, p, subt) Vec t[3]; Vec *p; Vec subt[3];
 *	In which of t[]'s facets does point p lie?  Return facet number
 *	and store its vertices into "subt".
 *
 * Higher-level interface:
 * int name2facet(name, verts) char *name; Vec verts[3];
 *	Interpret a complete triangle code name; store vertices into verts[].
 * void point2name(p, name, depth, verts)
 *			Vec *p; char name[]; int depth; Vec *verts;
 *	Find which triangle, to subdivision level 'depth', contains point p.
 *
 * I/O conversion:
 * int whatsign(ch) int ch;
 * double dms2rad(dms) char *dms;
 *	Parse degrees/minutes/seconds with optional +/-/N/S/E/W sign.
 * char *rad2dms(rad, suffix) double rad; char *suffix;
 *	Convert angle in radians into degrees/minutes/seconds with sign.
 * void geo2point(lat, lon, point) double lat, lon; Vec *point;
 *	Convert latitude/longitude to 3-D point on the unit sphere.
 * void point2geo(point, latp, lonp) Vec *point; double *latp, *lonp;
 *	Convert 3-D point to latitude/longitude in radians.
 * char *point2str(p) Vec *p;
 *	Return formatted string representing a 3-D point's lat/lon.
 */

/*
 * We really need double, not float, precision -- 24-bit mantissa is too small.
 * If floating-point is troublesome, 32-bit fixed-point, with binary point
 * chosen to allow values in the range -3.999 .. 3.999, might just do it.
 */ 


typedef struct vec {
    double x, y, z;
} Vec;

char *point2str();

/***************************** Basic geometry ********************************/

/*
 * Compute dot product between two vectors.
 */
double
dot(v1, v2)
    Vec *v1, *v2;
{
    return v1->x * v2->x + v1->y * v2->y + v1->z * v2->z;
}

/*
 * Great-circle distance between two points.
 */
double
distance(v1, v2)
    Vec *v1, *v2;
{
    double dx = v1->x - v2->x;
    double dy = v1->y - v2->y;
    double dz = v1->z - v2->z;
    double s2 = (dx*dx + dy*dy + dz*dz) / 4;
    return 2*atan2(sqrt(s2), sqrt(1-s2));
}

void
cross(v1, v2, result)
    Vec *v1, *v2, *result;
{
    result->x = v1->y*v2->z - v1->z*v2->y;
    result->y = v1->z*v2->x - v1->x*v2->z;
    result->z = v1->x*v2->y - v1->y*v2->x;
}

/*
 * Normalize a vector -- scale it to lie on the unit sphere.
 */
double
normalize(v)
    Vec *v;
{
    double s = sqrt(dot(v,v));
    if(s > 0) {
	v->x /= s; v->y /= s; v->z /= s;
    }
    return s;
}

/* Linear combination of vectors:  result = v1 + s*v2 */
void
comb(v1, s, v2, result)
    Vec *v1, *v2, *result;
    double s;
{
    result->x = v1->x + s*v2->x;
    result->y = v1->y + s*v2->y;
    result->z = v1->z + s*v2->z;
}

/* Build a basis for the tangent plane at p.
 * Set (vx,vy) perpendicular to p, with vy aligned with up.
 * Assumes p already normalized.
 */
basis(p, up, vx, vy)
    Vec *p, *up, *vx, *vy;
{
    double s = -dot(p, up);
    comb(up, s, p, vy);
    if(normalize(vy) == 0.)
	vy->y = 1;
    cross(p, vy, vx);
    normalize(vx);
}

/* Interpolate between three vectors,
 * and push the result back to the sphere
 */
void
interp(v1, v2, v3, result)
    Vec *v1, *v2, *v3, *result;
{
    result->x = v1->x + v2->x + v3->x;
    result->y = v1->y + v2->y + v3->y;
    result->z = v1->z + v2->z + v3->z;
    normalize(result);
}

/* On which side of the great circle (va,vb) does point v lie?
 * Computed as (va cross vb) dot v.
 * Note we use >= 0, to ensure we're on the "right side" of both (va,vb)
 * and (vb,va) if we happen to lie just on the line.
 */
int
rightside(v, va, vb)
    Vec *v, *va, *vb;
{
    Vec axb;

    cross(va, vb, &axb);
    return dot(&axb, v) >= 0;
}


/*
 * Does point v lie within triangle (va, vb, vc)?
 */
int
within(v, va, vb, vc)
    Vec *v, *va, *vb, *vc;
{
    return rightside(v,va,vb) && rightside(v,vb,vc) && rightside(v,vc,va);
}

#ifdef CHECK

/* Returns distance(v, line(va,vb)).
 * Used for checking the algorithm.  If a point should lie within some
 * triangle, but testing shows it lies slightly outside due to numerical
 * error, this routine can push it back in.
 * If v lies leftward of (va,vb), this routine sets *movedv to the
 * nearest point to v lying on (va,vb).
 */
double
rightby(v, va, vb, movedv)
    Vec *v, *va, *vb, *movedv;
{
    Vec axb;
    double d;
    cross(va,vb, &axb);
    normalize(&axb);
    d = dot(v, &axb);
    if(d < 0)
	comb(v, -d, &axb, movedv);
    return d;
}

double
withinby(v, va, vb, vc)
    Vec *v, *va, *vb, *vc;
{
    
    double dist = rightby(v,va,vb, v);
    double d;
    d = rightby(v,vb,vc, v);
    if(d < dist) dist = d;
    d = rightby(v,vc,va, v);
    if(d < dist) dist = d;
    return dist;
}
#endif /*CHECK*/

/********************* Subdivision routines ******************************/

/*
 * Subdivision pattern:
 *
 * The plane triangle (v0,v1,v2) is subdivided uniformly in thirds, then all
 * vertices are raised to lie on the unit sphere.
 * So e.g. v3 = normalized( (2/3)*v0 + (1/3)*v1 ).
 * We could instead have subdivided uniformly on the sphere itself, but doing
 * it this way actually yields triangles that are more-nearly equilateral.
 *
 * Here's the labelling for subdivided vertices and triangles:

	    v0
	   /  \
	  /  9 \
	 v3----v8
	/ \  8 / \
       / 4 \  / 7 \
      v4----v9----v7
     / \ 3 /  \ 6 / \
    / 1 \ /  2 \ / 5 \
   v1----v5----v6----v2

 */

int subvert[10-3][3] = { /* Which corners to interpolate for each sub-vertex? */
   { 0,0,1 },	/* v3 */
   { 0,1,1 },	/* v4 */
   { 1,1,2 },	/* v5 */
   { 1,2,2 },	/* v6 */
   { 0,2,2 },	/* v7 */
   { 0,0,2 },	/* v8 */
   { 0,1,2 }	/* v9 */
};

int subface[10][3] = {   /* Which vertices make up each subdivided face? */
  { 0,1,2 },	/* Not really used */
  { 4,1,5 },	/* facet 1 */
  { 9,5,6 },	/* facet 2 */
  { 5,9,4 },	/* facet 3 */
  { 3,4,9 },	/* facet 4 */
  { 7,6,2 },	/* facet 5 */
  { 6,7,9 },	/* facet 6 */
  { 8,9,7 },	/* facet 7 */
  { 9,8,3 },	/* facet 8 */
  { 0,3,8 }	/* facet 9 */
};

/*
 * Subdivide a triangle -- return in subt[] the vertices of the
 * facet'th sub-triangle of triangle tri[].
 */
void
subtri(tri, facet, subt)
    Vec tri[3];		/* Original triangle */
    int facet;		/* Face number 1..9 */
    Vec subt[3];	/* Subdivided triangle; may be same memory as tri[] */
{
    int i, k;
    Vec t[3];

    t[0] = tri[0];  t[1] = tri[1];  t[2] = tri[2];
    for(i = 0; i < 3; i++) {
	k = subface[facet][i];
	if(k < 3)
	    subt[i] = t[k];
	else
	    interp(&t[subvert[k-3][0]], &t[subvert[k-3][1]],
		   &t[subvert[k-3][2]], &subt[i]);
    }
}

/*
 * Assuming that triangle t[] contains point p,
 * in which of t[]'s facets does it lie?
 * Return the facet number, and leave the facet's vertices in subt[].
 */
int
findtri(t, p, subt)
    Vec t[3];
    Vec *p;
    Vec subt[3];
{
    Vec v[10];
    int i, k, facet;

    /* Fill v[] with the 10 vertices of the subdivision */
    v[0] = t[0];  v[1] = t[1];  v[2] = t[2];

    for(i = 0; i < 10-3; i++) {
	interp(&t[subvert[i][0]], &t[subvert[i][1]], &t[subvert[i][2]],
		&v[i+3]);
    }

    /* This decision tree discovers in which sub-facet point "p" lies.
     * We take advantage of the fact that edges which look straight in the
     * subdivision diagram are actually straight, since we're doing
     * linear interpolation on the plane (the flat triangle between the
     * vertices).  So we get subdivision points which are collinear on the
     * plane; when raised to the sphere, those points lie on the same great
     * circle.  This wouldn't be true if, say, the subdivision rule divided
     * each great-circle arc into equal pieces.  We go for equal chords,
     * not equal angles.
     */

	/* "Is p on the right-hand side of the line from vertex a to b?" */

#define RIGHT(a,b)  rightside(p, &v[a], &v[b])		/* shorthand */

    if(RIGHT(6,3)) {
	if(RIGHT(9,5)) facet = 2;
	else if(RIGHT(5,4)) facet = 1;
	else if(RIGHT(4,9)) facet = 4;
	else facet = 3;
    } else {
	if(RIGHT(7,9)) {
	    if(RIGHT(6,7)) facet = 6;
	    else facet = 5;
	} else {
	    if(RIGHT(8,9)) facet = 7;
	    else if(RIGHT(8,3)) facet = 8;
	    else facet = 9;
	}
    }

    for(i = 0; i < 3; i++)
	subt[i] = v[subface[facet][i]];

#ifdef CHECK

/*
 * Check that our point really lies within the sub-facet we think it does.
 * This shouldn't really be necessary in production code.
 * It's possible that it lies just barely outside our triangle,
 * due to numerical error.
 */
    if(!within(p, &subt[0], &subt[1], &subt[2])) {
	fprintf(stderr, "findtri: missed by %g!  Thought %d\n",
		withinby(p, &subt[0], &subt[1], &subt[2]), facet);
	fprintf(stderr, "p %s\n", point2str(p));
	for(i=0;i<3; i++) {
	    fprintf(stderr, " p%d=v%d %s\n", i, subface[facet][i],
			point2str(&subt[i]));
	}
	fprintf(stderr, "\n RS(p0,p1) %d RS(p1,p2) %d RS(p2,p0) %d\n",
		rightside(p,&subt[0],&subt[1]),
		rightside(p,&subt[1],&subt[2]),
		rightside(p,&subt[2],&subt[0]));
	return -1;
    }
#endif /*CHECK*/

    return facet;
}

/*
 * These two tables define our icosahedral triangulation of the sphere.
 */

Vec vico[12] = {
 { 0,			    0,			     1.},		  /*0*/
 { 0.89442719099991587856,  0,			     0.44721359549995793},/*1*/
 { 0.27639320225002104342,  0.85065080835203993366,  0.44721359549995793},/*2*/
 {-0.72360679774997893378,  0.52573111211913365982,  0.44721359549995793},/*3*/
 {-0.72360679774997893378, -0.52573111211913365982,  0.44721359549995793},/*4*/
 { 0.27639320225002104342, -0.85065080835203993366,  0.44721359549995793},/*5*/
 { 0.72360679774997893378,  0.52573111211913365982, -0.44721359549995793},/*6*/
 {-0.27639320225002104342,  0.85065080835203993366, -0.44721359549995793},/*7*/
 {-0.89442719099991587856,  0,			    -0.44721359549995793},/*8*/
 {-0.27639320225002104342, -0.85065080835203993366, -0.44721359549995793},/*9*/
 { 0.72360679774997893378, -0.52573111211913365982, -0.44721359549995793},/*10*/
 { 0,			    0,			    -1.}		  /*11*/
};

int fico[20][3] = {
 { 2, 0, 1 },		/* A */
 { 3, 0, 2 },		/* B */
 { 4, 0, 3 },		/* C */
 { 5, 0, 4 },		/* D */
 { 1, 0, 5 },		/* E */
 { 1, 6, 2 },		/* F */
 { 7, 2, 6 },		/* G */
 { 2, 7, 3 },		/* H */
 { 8, 3, 7 },		/* I */
 { 3, 8, 4 },		/* J */
 { 9, 4, 8 },		/* K */
 { 4, 9, 5 },		/* L */
 { 10, 5, 9 },		/* M */
 { 5, 10, 1 },		/* N */	/* N exchanged with O, 96.09.13, slevy */
 { 6, 1, 10 },		/* O */ /* to put equatorial row in consistent order. */
 { 6, 11, 7 },		/* P */
 { 7, 11, 8 },		/* Q */
 { 8, 11, 9 },		/* R */
 { 9, 11, 10 },		/* S */
 { 10, 11, 6 }		/* T */
};


/*
 * Given a triangle's name, e.g. "D83-29-7", find the coordinates
 * of its vertices (returned in verts[]).  Returns 1 if successful,
 * 0 if the name was invalid.
 */
int
name2facet(name, verts)
    char *name;
    Vec verts[3];
{
    int major = name[0];
    int i, facet;
    char *cp;

    if(major >= 'a' && major <= 't')
	major -= 'a';
    else if(major >= 'A' && major <= 'T')
	major -= 'A';
    else {
	fprintf(stderr, "%s: must begin with a letter A through T\n", name);
	return 0;
    }  

    for(i = 0; i < 3; i++)
	verts[i] = vico[fico[major][i]];
    
    for(cp = &name[1]; *cp != '\0';  cp++) {
	if(*cp >= '1' && *cp <= '9') {
	    facet = *cp - '0';
	    subtri(verts, facet, verts);
	} else if(*cp != ' ' && *cp != '-') {
	    fprintf(stderr,
		"%s: expected only digits 1..9, blanks, and hyphens\n", name);
	    return 0;
	}
    }
    return 1;
}

/*
 * Given a point, return the name of the facet it lies in (to depth 'depth'),
 * and the the facet's vertices (in verts[]).
 * name[] must have at least (3/2)*depth+2 bytes of room. 
 */
void
point2name(p, name, depth, verts)
    Vec *p;
    char name[];	/* Must have at least depth+2 characters */
    int depth;
    Vec *verts;
{
    char *cp;
    int i, facet;

    for(facet = 0; facet < 20; facet++) {
	for(i = 0; i < 3; i++)
	    verts[i] = vico[fico[facet][i]];
	if(within(p, &verts[0], &verts[1], &verts[2]))
	    break;
    }
    if(i >= 20) {
	fprintf(stderr, "Trouble: %g %g %g not within any triangle?\n",
		p->x, p->y, p->z);
	return;
    }

    name[0] = facet+'A';
    cp = &name[1];
    for(i = 0; i < depth; i++) {
	facet = findtri(verts, p, verts);
	if(facet < 0)
	    break;
	if(i%2 == 0 && i > 0)
	    *cp++ = '-';
	*cp++ = facet+'0';
    }
    *cp = '\0';	/* null-terminate */
}

/***************************** I/O routines *********************************/

int
whatsign(ch)
    int ch;
{
    switch(ch) {
    case '-': case 'w': case 'W': case 's': case 'S': return -1;
    case '+': case 'e': case 'E': case 'n': case 'N': return 1;
    default:					      return 0;
    }
}

double
dms2rad(dms)
    char *dms;
{
    char c1, c2;
    int sign;
    double d, m = 0, s = 0;
    char *str = dms;

    sign = whatsign(str[0]);
    if(sign != 0)
	str++;

    if(sscanf(str, "%lf%c%lf%c%lf", &d, &c1, &m, &c2, &s) <= 0) {
	fprintf(stderr, "Expected [-]dd:mm:ss[NSEW], can't interpret \"%s\"\n",
		dms);
	return 0;
    }

    /* Check for a suffix */
    if(sign == 0)
	sign = whatsign(dms[ strlen(dms) - 1 ]);
    if(sign == 0)
	sign = 1;

    return (sign * M_PI/180.0 * (d + (m + s/60.0)/60.0));
}

char *
rad2dms(rad, suffix)
    double rad;
    char *suffix;
{
    static char dms[32];
    double d, m;  int id, im;
    int s;
    char sign;

    sign = (rad < 0) ? suffix[1] : suffix[0];
    /* Convert to degrees, rounded to nearest .01 second,
     * since that's what we're printing below.
     * Rounding prevents oddities like "89:59:60.00"
     */
    d = fabs(rad)*180.0/M_PI + 0.005 / 3600.0;
    m = (d - floor(d)) * 60.0;
    s = (m - floor(m)) * 6000.0;
    id = d;
    im = m;
    sprintf(dms, "%3d:%02d:%02d.%02d%c", id, im, s/100, s%100, sign);
    return (dms);
}

char *
rad2deg(rad, suffix)
    double rad;
    char *suffix;
{
    static char deg[32];
    double d;
    char sign;

    sign = (rad < 0) ? suffix[1] : suffix[0];
    /* Convert to degrees, rounded to nearest .01 second,
     * since that's what we're printing below.
     * Rounding prevents oddities like "89:59:60.00"
     */
    d = fabs(rad)*180.0/M_PI;
    sprintf(deg, "%12.7f%c", d, sign);
    return (deg);
}

void
geo2point(lat, lon, point)
    double lat, lon;
    Vec *point;
{
    double r;
    r = cos(lat);
    point->x = r*cos(lon);
    point->y = r*sin(lon);
    point->z = sin(lat);
}

void
point2geo(point, latp, lonp)
    Vec *point;
    double *latp, *lonp;
{
    double z, r;
    r = sqrt(point->x*point->x + point->y*point->y);
    *latp = atan2(point->z, r);
    if(r == 0) *lonp = 0;
    else *lonp = atan2(point->y, point->x);
}

enum { XYZ, DMS, DEG };

int ll = DMS;

char *
point2str(p)
    Vec *p;
{
    double lat, lon;
    static char str[64];
    switch(ll) {
    case XYZ:
	sprintf(str, "%.16g %.16g %.16g", p->x,p->y,p->z);
	break;
    case DMS:
	point2geo(p, &lat, &lon);
	sprintf(str, "%s ", rad2dms(lat, "NS"));
	sprintf(str+strlen(str), "%s", rad2dms(lon, "EW"));
	break;
    case DEG:
	point2geo(p, &lat, &lon);
	sprintf(str, "%s ", rad2deg(lat, "NS"));
	sprintf(str+strlen(str), "%s", rad2deg(lon, "EW"));
	break;
    }
    return str;
}

double
bearing(p1, p2)
    Vec *p1, *p2;
{
    Vec vx, vy, disp;
    static Vec up = { 0,0,1 };
    double b;
    basis(p1, &up, &vx, &vy);
    comb(p2, -1., p1, &disp);	/* disp = p2 - p1 */
    b = -atan2(dot(&disp, &vx), dot(&disp, &vy));
    if(b < 0) b += 2*M_PI;
    return b;
}

/*
 * With one argument, converts a triangle name to latitude/longitude.
 * With two arguments, converts lat/lon to a triangle.
 */

#define Rearth  6371e3		/* Earth's mean radius is about 6371 km. */
#define	DEPTH	16		/* Number of digits in our triangle codes */
				/* 16 digits yields approx. 20 cm resolution */

char Usage1[] = "\
Usage:  %s <triangle-code>   -or-   %s <latitude>  <longitude>\n\
Given a triangle code, computes the latitude & longitude of the center\n\
and corners of the triangle, and the triangle's edge lengths.\n\
Given latitude and longitude, finds the triangle containing that point.\n\
Triangle codes begin with a letter from A to T, followed by a string of\n\
digits from 1 through 9 (not 0).  Optional hyphens are ignored.\n";

char Usage2[] = "\
Latitude/longitude are specified in degrees, or degrees:minutes:seconds,\n\
with optional N/S/E/W prefix or suffix.  If presented as a signed number,\n\
latitude is positive when north, longitude positive when east.\n\
Examples:\n\
	%s  d83-53-31-22-82-21\n\
	%s  44:58:37.5  W93:20:07\n\
	%s  44.977 -93.335\n";


readlines(inf, prompt)
    FILE *inf;
    char *prompt;
{
    char line[128];
    char *argv[8];
    int i, status = 1;
    char *p;

    argv[0] = "subdiv";
    if(prompt) {
	printf("%s", prompt);
	fflush(stdout);
    }
    while(fgets(line, sizeof(line), inf) != NULL) {
    	for(i = 1, p = line; i < 8; i++) {
    	    while(*p == ' ' || *p == '\t' || *p == '\n') p++;
    	    argv[i] = p;
    	    if(*p == '\0') break;
    	    while(*p != ' ' && *p != '\t' && *p != '\n' && *p) p++;
    	    if(*p == '\0') break;
    	    *p++ = '\0';
    	}
    	argv[i] = NULL;
    	if(i <= 1) break;
	status = processargs(i, argv);
	if(prompt) {
	    printf("%s", prompt);
	    fflush(stdout);
	}
    }
    return status;
}

int quick = 0;
int duplicate = 0;

#if !( defined(__unix__) || defined(__APPLE__) || defined(__linux__) )
main()
{
    return readlines(stdin, "<triangle-code -or- lat[NS] lon[EW]>: ");
}

#else /* unix */

main(argc, argv)
    int argc;
    char *argv[];
{
    return processargs(argc, argv);
}
#endif


processargs(argc, argv)
    int argc;
    char *argv[];
{
    double lat, lon;
    Vec point, point2, tri[3], center, nearby;
    char name[2*DEPTH+2];
    char *prog = argv[0];

    while(argc > 1 && argv[1][0] == '-' && isalpha(argv[1][1])) {
	if(strcmp(argv[1],"-p")==0)
	    ll = XYZ;
	else if(strcmp(argv[1],"-d")==0)
	    ll = DEG;
	else if(strcmp(argv[1],"-dms")==0)
	    ll = DMS;
	else if(strcmp(argv[1],"-q")==0)
	    quick = 1;
	else if(strcmp(argv[1],"-c")==0)
	    quick = -1;			/* Center only */
	else if(strcmp(argv[1], "-span")==0)
	    duplicate = 1;
	else if(strcmp(argv[2],"-r")==0) {
	    return readlines(stdin, NULL);
	} else {
	    fprintf(stderr, "Usage: %s [-p|-d|-dms] [-q] [-span] [-r] {triname|lat lon}\n\
Convert between triangle-codes and latitude/longitude locations.\n\
-p : print locations as X Y Z (on unit sphere)\n\
-d : print locations as decimal degrees\n\
-dms : print locations as degrees:minutes:seconds\n\
-q : quick -- just print one-line triangle->location code in decimal degrees\n\
-span : duplicate triangles which span 180-degree longitude line.\n", prog);
	    return(1);
	}
	argc--, argv++;
    }


    switch(argc-1) {
    case 2:		/* two arguments: latitude and longitude */
	lat = dms2rad(argv[1]);
	lon = dms2rad(argv[2]);
	geo2point(lat, lon, &point);

	point2name(&point, name, DEPTH, tri);

	printf("%s\n", name);
	interp(&tri[0], &tri[1], &tri[2], &center);
	printf("Check: center %s\n", point2str(&center));
	break;

    case 1:		/* one argument: triangle name */
	if( name2facet(argv[1], tri) == 0)
	    return(1);
	interp(&tri[0], &tri[1], &tri[2], &center);
	if(quick) {
		double lat[4], lon[4], dlon;
		int shoved = 0;
		static char fmt[] = "%s\t%.3f %.3f\t%.3f %.3f\t%.3f %.3f\n";
		static char cfmt[] = "%s\t%.3f %.3f\n";
		point2geo(&tri[0], &lat[0], &lon[0]);
		point2geo(&tri[1], &lat[1], &lon[1]);
		point2geo(&tri[2], &lat[2], &lon[2]);
		point2geo(&center, &lat[3], &lon[3]);
		dlon = (lon[3] > 0) ? 2*M_PI : -2*M_PI;
		if(lon[0]*lon[3] < 0)
			lon[0] += dlon, shoved = 1;
		if(lon[1]*lon[3] < 0)
			lon[1] += dlon, shoved = 1;
		if(lon[2]*lon[3] < 0)
			lon[2] += dlon, shoved = 1;
		if(quick < 0) {
		    printf(cfmt, argv[1], R2D(lat[3]), R2D(lon[3]));
		} else {
		    printf(fmt, argv[1],
			R2D(lat[0]), R2D(lon[0]),
			R2D(lat[1]), R2D(lon[1]),
			R2D(lat[2]), R2D(lon[2]));
		}
		if(shoved && duplicate) {
		    if(quick < 0) {
			printf(cfmt, argv[1], R2D(lat[3]), R2D(lon[3]-dlon));
		    } else {
			printf(fmt, argv[1],
			    R2D(lat[0]), R2D(lon[0]-dlon),
			    R2D(lat[1]), R2D(lon[1]-dlon),
			    R2D(lat[2]), R2D(lon[2]-dlon));
		    }
		}
		break;
	}

	printf("center %s  center->1  %.3f m at %6.2f degrees\n",
		point2str(&center), Rearth*distance(&center, &tri[1]),
		180/M_PI * bearing(&center, &tri[1]));
	printf("vert 0 %s  side 1->2  %.3f m at %6.2f degrees\n",
		point2str(&tri[0]),
		Rearth*distance(&tri[1],&tri[2]),
		180/M_PI * bearing(&tri[1], &tri[2]));
	printf("vert 1 %s  side 0->2  %.3f m at %6.2f degrees\n",
		point2str(&tri[1]),
		Rearth*distance(&tri[0],&tri[2]),
		180/M_PI * bearing(&tri[0], &tri[2]));
	printf("vert 2 %s  side 0->1  %.3f m at %6.2f degrees\n",
		point2str(&tri[2]),
		Rearth*distance(&tri[0],&tri[1]),
		180/M_PI * bearing(&tri[0], &tri[1]));

	/* Pick a new point, slightly displaced toward vertex 0,
	 * and find its triangle code as a check.
	 * We could just ask about the center itself, but that tends to
	 * yield complaints about numerical error, as the center lies
	 * on the border of six triangles.
	 */
	interp(&tri[0], &center, &center, &nearby);
	point2name(&nearby, name, DEPTH, tri);
	printf("Check: %s\n", name);
#ifdef DEBUG
	point2name(&center, name, DEPTH, tri);
	printf("Check2: %s\n", name);
#endif
	break;

    case 4:	/* Four arguments: find bearing between two points */
		/* This option isn't advertised in the Usage message. */
	geo2point(dms2rad(argv[1]), dms2rad(argv[2]), &point);
	geo2point(dms2rad(argv[3]), dms2rad(argv[4]), &point2);
	printf("%.3f m at bearing %6.2f degrees\n",
		distance(&point, &point2) * Rearth,
		bearing(&point, &point2) * 180/M_PI);
	break;
	
    default:
	fprintf(stderr, Usage1, argv[0], argv[0]);
	fprintf(stderr, Usage2, argv[0], argv[0], argv[0]);
	return(1);
    }
    return(0);
}
